#pragma once

#include <cassert>
#include <cstdint>
#include <cstring>
#include <exception>
#include <fstream>
#include <string>

#include <minimp4.h>

#include <wels/codec_api.h>


class VideoEncoder final {
 public:
  VideoEncoder() = delete;
  VideoEncoder(const std::string& path, const SEncParamBase& p) :
      file_(path, std::ofstream::binary), fps_(p.fMaxFrameRate) {
    if (WelsCreateSVCEncoder(&encoder_)) {
      throw std::runtime_error {"failed to init openh264 encoder"};
    }

    muxer_ = MP4E_open(false, false, this, WriteCallback);
    if (MP4E_STATUS_OK != mp4_h26x_write_init(&writer_, muxer_, p.iPicWidth, p.iPicHeight, false)) {
      throw std::runtime_error {"failed to init h26x writer"};
    }

    encoder_->Initialize(&p);

    int lv = WELS_LOG_DEBUG;
    encoder_->SetOption(ENCODER_OPTION_TRACE_LEVEL, &lv);
    int fmt = videoFormatI420;
    encoder_->SetOption(ENCODER_OPTION_DATAFORMAT, &fmt);
  }
  ~VideoEncoder() noexcept {
    encoder_->Uninitialize();
    WelsDestroySVCEncoder(encoder_);

    MP4E_close(muxer_);
    mp4_h26x_write_close(&writer_);
  }

  void Encode(const SSourcePicture& pic) {
    SFrameBSInfo info;
    if (cmResultSuccess != encoder_->EncodeFrame(&pic, &info)) {
      throw std::runtime_error {"failed to encode a frame"};
    }
    if (info.eFrameType == videoFrameTypeSkip) return;

    for (size_t i = 0; i < static_cast<size_t>(info.iLayerNum); ++i) {
      const auto& lay = info.sLayerInfo[i];

      uint8_t* buf = lay.pBsBuf;
      for (size_t j = 0; j < static_cast<size_t>(lay.iNalCount); ++j) {
        mp4_h26x_write_nal(&writer_, buf, lay.pNalLengthInByte[j], static_cast<unsigned int>(90000./fps_));
        buf += lay.pNalLengthInByte[j];
      }
    }
  }

 private:
  std::ofstream file_;

  ISVCEncoder*      encoder_;
  MP4E_mux_t*       muxer_;
  mp4_h26x_writer_t writer_;

  double fps_;


  static int WriteCallback(int64_t off, const void* buf, size_t size, void* ptr) noexcept {
    auto self = (VideoEncoder*) ptr;
    self->file_.seekp(off);
    assert(self->file_);
    self->file_.write(static_cast<const char*>(buf), static_cast<std::streamsize>(size));
    return !self->file_;
  }
};
